import logging
from typing import Optional, Dict, Type

import slixmpp.stanza
from slixmpp.plugins.xep_0004 import Form

import spade.message

SPADE_X_METADATA = "spade:x:metadata"

logger = logging.getLogger("spade.Message")


class MessageBase(object):
    """ """

    def __init__(
        self,
        to: Optional[str] = None,
        sender: Optional[str] = None,
        body: Optional[str] = None,
        thread: Optional[str] = None,
        metadata: Optional[Dict[str, str]] = None,
    ):
        self.sent = False
        self._to, self._sender, self._body, self._thread = None, None, None, None
        self.to = to
        self.sender = sender
        self.body = body
        self.thread = thread

        if metadata is None:
            self.metadata = {}
        else:
            for key, value in metadata.items():
                if not isinstance(key, str) or not isinstance(value, str):
                    raise TypeError("Key and Value of metadata MUST be strings")
            self.metadata = metadata

    @classmethod
    def from_node(cls, node: slixmpp.Message) -> Type["MessageBase"]:
        """
        Creates a new spade.message.Message from a slixmpp.stanza.Message

        Args:
          node (slixmpp.stanza.Message): a slixmpp Message

        Returns:
          spade.message.Message: a new spade Message

        """
        if not isinstance(node, slixmpp.stanza.Message):
            raise AttributeError("node must be a slixmpp.stanza.Message instance")
        msg = cls()
        msg._to = node["to"]
        msg._sender = node["from"]

        if isinstance(node["body"], dict):
            for body in node["body"].values():
                msg.body = body
                break
        else:
            msg.body = node["body"]

        for data in [pl for pl in node.get_payload() if pl.tag == "{jabber:x:data}x"]:
            if data.find("{jabber:x:data}title").text == SPADE_X_METADATA:
                for field in data.findall("{jabber:x:data}field"):
                    if field.attrib["var"] != "_thread_node":
                        msg.set_metadata(
                            field.attrib["var"], field.find("{jabber:x:data}value").text
                        )
                    else:
                        msg.thread = field.find("{jabber:x:data}value").text

        return msg

    @property
    def to(self) -> slixmpp.JID:
        """
        Gets the jid of the receiver.

        Returns:
          slixmpp.JID: jid of the receiver

        """
        return self._to

    @to.setter
    def to(self, jid: str) -> None:
        """
        Set jid of the receiver.

        Args:
          jid (str): the jid of the receiver.

        """
        if jid is not None and not isinstance(jid, str):
            raise TypeError("'to' MUST be a string")
        self._to = slixmpp.JID(jid) if jid is not None else None

    @property
    def sender(self) -> slixmpp.JID:
        """
        Get jid of the sender

        Returns:
          slixmpp.JID: jid of the sender

        """
        return self._sender

    @sender.setter
    def sender(self, jid: str) -> None:
        """
        Set jid of the sender

        Args:
          jid (str): jid of the sender

        """
        if jid is not None and not isinstance(jid, str):
            raise TypeError("'sender' MUST be a string")
        self._sender = slixmpp.JID(jid) if jid is not None else None

    @property
    def body(self) -> str:
        """
        Get body of the message
        Returns:
            str: the body of the message
        """
        return self._body

    @body.setter
    def body(self, body: str) -> None:
        """
        Set body of the message
        Args:
            body (str): The body of the message
        """
        if body is not None and not isinstance(body, str):
            raise TypeError("'body' MUST be a string")
        self._body = body

    @property
    def thread(self) -> str:
        """
        Get Thread of the message

        Returns:
            str: thread id
        """
        return self._thread

    @thread.setter
    def thread(self, value: str) -> None:
        """
        Set thread id of the message

        Args:
            value (str): the thread id

        """
        if value is not None and not isinstance(value, str):
            raise TypeError("'thread' MUST be a string")
        self._thread = value

    def set_metadata(self, key: str, value: str) -> None:
        """
        Add a new metadata to the message

        Args:
          key (str): name of the metadata
          value (str): value of the metadata

        """
        if not isinstance(key, str) or not isinstance(value, str):
            raise TypeError("'key' and 'value' of metadata MUST be strings")
        self.metadata[key] = value

    def get_metadata(self, key: str) -> str:
        """
        Get the value of a metadata. Returns None if metadata does not exist.

        Args:
          key (str): name of the metadata

        Returns:
          str: the value of the metadata (or None)

        """
        return self.metadata[key] if key in self.metadata else None

    def match(self, message: Type["MessageBase"]) -> bool:
        """
        Returns wether a message matches with this message or not.
        The message can be a Message object or a Template object.

        Args:
          message (spade.message.Message): the message to match to

        Returns:
          bool: wether the message matches or not

        """
        if self.to and message.to != self.to:
            return False

        if self.sender and message.sender != self.sender:
            return False

        if self.body and message.body != self.body:
            return False

        if self.thread and message.thread != self.thread:
            return False

        for key, value in self.metadata.items():
            if message.get_metadata(key) != value:
                return False

        logger.debug(f"message matched {self} == {message}")
        return True

    @property
    def id(self) -> int:
        """ """
        return id(self)

    def __eq__(self, other: Type["MessageBase"]):
        if type(other) is not spade.message.Message:
            return False
        return self.match(other)


class Message(MessageBase):
    """ """

    def make_reply(self) -> "Message":
        """
        Creates a copy of the message, exchanging sender and receiver

        Returns:
          spade.message.Message: a new message with exchanged sender and receiver

        """
        return Message(
            to=str(self.sender),
            sender=str(self.to),
            body=self.body,
            thread=self.thread,
            metadata=self.metadata,
        )

    def prepare(self) -> slixmpp.stanza.Message:
        """
        Returns a slixmpp.stanza.Message built from the Message and prepared to be sent.

        Returns:
          slixmpp.stanza.Message: the message prepared to be sent

        """
        msg = slixmpp.stanza.Message()
        msg["to"] = self.to
        msg["from"] = self.sender
        msg["body"] = self.body
        msg.chat()

        # Send metadata using xep-0004: Data Forms (https://xmpp.org/extensions/xep-0004.html)
        if len(self.metadata):
            form = Form()
            form["type"] = "form"
            for name, value in self.metadata.items():
                form.add_field(var=name, ftype="text-single", value=value)

            if self.thread:
                form.add_field(
                    var="_thread_node", ftype="text-single", value=self.thread
                )

            form["title"] = SPADE_X_METADATA
            msg.append(form)

        return msg

    def __str__(self) -> str:
        s = f'<message to="{self.to}" from="{self.sender}" thread="{self.thread}" metadata={self.metadata}>'
        if self.body:
            s += "\n" + self.body + "\n"
        s += "</message>"
        return s



# -*- coding: utf-8 -*-

"""The setup script."""
import sys
from setuptools import setup, find_packages


def parse_requirements(filename):
    """ load requirements from a pip requirements file """
    with open(filename) as f:
        lineiter = [line.strip() for line in f]
    return [line for line in lineiter if line and not line.startswith("#")]


with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

requirements = parse_requirements("requirements.txt")

if sys.platform.startswith("win32"):
    requirements.append("winloop>=0.1.7")
else:
    requirements.append("uvloop>=0.21.0")

setup_requirements = [
    'pytest-runner',
    # put setup requirements (distutils extensions, etc.) here
]

test_requirements = parse_requirements("requirements_dev.txt")

setup(
    name='spade',
    version='4.0.0rc2',
    description="Smart Python Agent Development Environment",
    long_description=readme + '\n\n' + history,
    author="Javi Palanca",
    author_email='jpalanca@gmail.com',
    url='https://github.com/javipalanca/spade',
    packages=find_packages(include=['spade']),
    entry_points={
        'console_scripts': [
            'spade=spade.cli:cli'
        ]
    },
    include_package_data=True,
    package_data={"spade": ["templates"]},
    install_requires=requirements,
    license="MIT license",
    zip_safe=False,
    keywords='spade',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: POSIX :: Linux',
        'Operating System :: Microsoft :: Windows',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: Internet :: XMPP',
    ],
    test_suite='tests',
    tests_require=test_requirements,
    setup_requires=setup_requirements,
)



from enum import Enum
from typing import Dict, Optional, Union

from slixmpp import JID
from slixmpp.stanza import Presence


class ContactNotFound(Exception):
    pass


class PresenceNotFound(Exception):
    pass


class PresenceShow(Enum):
    EXTENDED_AWAY = "xa"
    AWAY = "away"
    CHAT = "chat"
    DND = "dnd"
    NONE = "none"


class PresenceType(Enum):
    AVAILABLE = "available"
    UNAVAILABLE = "unavailable"
    ERROR = "error"
    PROBE = "probe"
    SUBSCRIBE = "subscribe"
    SUBSCRIBED = "subscribed"
    UNSUBSCRIBE = "unsubscribe"
    UNSUBSCRIBED = "unsubscribed"


class PresenceInfo:
    def __init__(
        self,
        presence_type: PresenceType,
        show: PresenceShow,
        status: Optional[str],
        priority: int,
    ):
        self.type = presence_type
        self.show = show
        self.status = status
        self.priority = priority

    def is_available(self) -> bool:
        return self.type == PresenceType.AVAILABLE

    def __eq__(self, other):
        if not isinstance(other, PresenceInfo):
            return False
        return (
            self.type == other.type
            and self.show == other.show
            and self.status == other.status
            and self.priority == other.priority
        )

    def __ne__(self, other):
        return not self.__eq__(other)

    def __str__(self):
        return f"PresenceInfo(Type: {self.type}, Show: {self.show}, Status: {self.status}, Priority: {self.priority})"

    def __repr__(self) -> str:
        return str(self)


class Contact:
    def __init__(self, jid: JID, name: str, subscription: str, ask: str, groups: list):
        self.jid = jid
        self.name = name
        self.subscription = subscription
        self.ask = ask
        self.groups = groups
        self.resources: Dict[str, PresenceInfo] = {}
        self.current_presence: Optional[PresenceInfo] = None
        self.last_presence: Optional[PresenceInfo] = None

    def update_presence(self, resource: str, presence_info: PresenceInfo):
        # Update the current presence automatically based on priority
        if presence_info == self.resources.get(resource):
            return
        self.resources[resource] = presence_info
        new_presence = max(
            self.resources.values(), key=lambda p: p.priority, default=None
        )
        if new_presence != self.current_presence:
            self.last_presence = self.current_presence
            self.current_presence = new_presence

    def get_presence(self, resource: Optional[str] = None) -> PresenceInfo:
        if resource:
            if resource in self.resources:
                return self.resources[resource]
            else:
                raise KeyError(
                    f"Resource '{resource}' not found for contact {self.jid}."
                )
        if self.current_presence:
            return self.current_presence
        raise PresenceNotFound(
            f"No presence information available for contact {self.jid}."
        )

    def update_subscription(self, subscription: str, ask: str):
        self.subscription = subscription
        self.ask = ask

    def is_available(self) -> bool:
        return (
            self.current_presence is not None
            and self.current_presence.type == PresenceType.AVAILABLE
        )

    def is_subscribed(self) -> bool:
        return self.subscription in ["both", "to"] or self.ask == "subscribe"

    def __str__(self):
        return f"Contact(JID: {self.jid}, Name: {self.name}, Presence: {self.current_presence})"

    def __repr__(self) -> str:
        return str(self)


class PresenceManager:
    def __init__(self, agent, approve_all: bool = False):
        self.contacts: Dict[str, Contact] = {}
        self.agent = agent
        self.current_presence: Optional[PresenceInfo] = None
        self.approve_all = approve_all
        # Adding event handlers to handle incoming presence and subscription events
        self.agent.client.add_event_handler("presence_available", self.handle_presence)
        self.agent.client.add_event_handler(
            "presence_unavailable", self.handle_presence
        )
        self.agent.client.add_event_handler("changed_status", self.handle_presence)
        self.agent.client.add_event_handler(
            "presence_subscribe", self.handle_subscription
        )
        self.agent.client.add_event_handler(
            "presence_subscribed", self.handle_subscription
        )
        self.agent.client.add_event_handler(
            "presence_unsubscribe", self.handle_subscription
        )
        self.agent.client.add_event_handler(
            "presence_unsubscribed", self.handle_subscription
        )
        self.agent.client.add_event_handler("roster_update", self.handle_roster_update)

    def is_available(self) -> bool:
        return (
            self.current_presence is not None and self.current_presence.is_available()
        )

    def get_presence(self) -> PresenceInfo:
        return self.current_presence

    def get_show(self) -> PresenceShow:
        return (
            self.current_presence.show if self.current_presence else PresenceShow.NONE
        )

    def get_status(self) -> Optional[str]:
        return self.current_presence.status if self.current_presence else None

    def get_priority(self) -> int:
        return self.current_presence.priority if self.current_presence else 0

    def handle_presence(self, presence: Presence):
        jid = presence["from"]
        peer_jid = str(jid)
        bare_jid = jid.bare
        if bare_jid == self.agent.jid.bare:
            return
        resource = presence["from"].resource
        presence_type = presence["type"]
        # Normalise the value of `type` if it is a show
        if presence_type in [show.value for show in PresenceShow]:
            presence_type = PresenceType.AVAILABLE
        presence_type = PresenceType(presence_type)

        show = PresenceShow(presence.get("show", "none"))
        status = presence.get("status")
        priority = int(presence.get("priority", 0))

        name = presence.name if presence.name else peer_jid
        presence_info = PresenceInfo(presence_type, show, status, priority)
        if bare_jid not in self.contacts:
            # Create a new contact if it doesn't exist
            self.contacts[bare_jid] = Contact(
                jid=JID(peer_jid), name=name, subscription="", ask="", groups=[]
            )
        # Update the presence of the contact
        self.contacts[bare_jid].update_presence(resource, presence_info)
        # Call user-defined handler
        if presence_type == PresenceType.AVAILABLE:
            self.on_available(
                peer_jid, presence_info, self.contacts[bare_jid].last_presence
            )
        elif presence_type == PresenceType.UNAVAILABLE:
            self.on_unavailable(
                peer_jid, presence_info, self.contacts[bare_jid].last_presence
            )

    def handle_subscription(self, presence: Presence):
        peer_jid = presence["from"].bare
        subscription_type = presence["type"]
        ask = presence.get("ask", "none")

        if peer_jid not in self.contacts:
            # Create a new contact if it doesn't exist
            self.contacts[peer_jid] = Contact(
                jid=JID(peer_jid),
                name=peer_jid,
                subscription=subscription_type,
                ask=ask,
                groups=[],
            )
        else:
            # Update the subscription information
            self.contacts[peer_jid].update_subscription(subscription_type, ask)

        # Call user-defined handler or automatically approve if approve_all is True
        if subscription_type == "subscribe" and self.approve_all:
            self.approve_subscription(peer_jid)
            self.on_subscribe(peer_jid)
        elif subscription_type == "subscribe":
            self.on_subscribe(peer_jid)
        elif subscription_type == "subscribed":
            self.on_subscribed(peer_jid)
        elif subscription_type == "unsubscribe":
            self.on_unsubscribe(peer_jid)
        elif subscription_type == "unsubscribed":
            self.on_unsubscribed(peer_jid)

    def handle_roster_update(self, event):
        """Executed when the roster is received or updated."""

        roster = event["roster"]
        for item in roster:
            bare_jid = item.get_jid().bare
            name = item.get("name", bare_jid)
            subscription = item.get("subscription", "none")
            ask = item.get("ask", "none")
            groups = item.get_groups()

            # Storing contact information in the internal structure
            if bare_jid not in self.contacts:
                self.contacts[bare_jid] = Contact(
                    jid=bare_jid,
                    name=name,
                    subscription=subscription,
                    ask=ask,
                    groups=groups,
                )
            else:
                self.contacts[bare_jid].name = name
                self.contacts[bare_jid].subscription = subscription
                self.contacts[bare_jid].ask = ask
                self.contacts[bare_jid].groups = groups

    def get_contact_presence(
        self, jid: Union[str, JID], resource: Optional[str] = None
    ) -> PresenceInfo:
        if isinstance(jid, JID):
            jid = jid.bare
        else:
            jid = JID(jid).bare
        if jid in self.contacts:
            return self.contacts[jid].get_presence(resource)
        else:
            raise ContactNotFound(f"Contact with JID '{jid}' not found.")

    def get_contact(self, jid: Union[str, JID]) -> Contact:
        if isinstance(jid, JID):
            jid = jid.bare
        else:
            jid = JID(jid).bare
        if jid in self.contacts:
            return self.contacts[jid]
        else:
            raise ContactNotFound(f"Contact with JID '{jid}' not found.")

    def get_contacts(self) -> Dict[str, Contact]:
        return {jid: c for jid, c in self.contacts.items() if c.is_subscribed()}

    def set_presence(
        self,
        presence_type: PresenceType = PresenceType.AVAILABLE,
        show: PresenceShow = PresenceShow.CHAT,
        status: Optional[str] = "",
        priority: int = 0,
    ):
        # This method could be used to set the presence for the local user
        self.current_presence = PresenceInfo(presence_type, show, status, priority)
        # Send the presence stanza to the server
        self.agent.client.send_presence(
            ptype=presence_type.value,
            pshow=show.value,
            pstatus=status,
            ppriority=str(priority),
        )

    def set_available(self):
        # Method to set presence to available
        self.set_presence(PresenceType.AVAILABLE)

    def set_unavailable(self):
        # Method to set presence to unavailable
        self.set_presence(PresenceType.UNAVAILABLE, PresenceShow.NONE, None, 0)

    def subscribe(self, jid: str):
        # Logic to send a subscription request to a contact
        if jid not in self.contacts:
            self.contacts[jid] = Contact(
                jid=JID(jid), name=jid, subscription="to", ask="subscribe", groups=[]
            )
        else:
            self.contacts[jid].update_subscription("to", "subscribe")
        # Send the subscription stanza to the server
        self.agent.client.send_presence(pto=jid, ptype="subscribe")

    def unsubscribe(self, jid: str):
        # Logic to send an unsubscription request to a contact
        if jid in self.contacts:
            self.contacts[jid].update_subscription("unsubscribe", "unsubscribe")
        # Send the unsubscription stanza to the server
        self.agent.client.send_presence(pto=jid, ptype="unsubscribe")

    def approve_subscription(self, jid: str):
        # Logic to approve a subscription request
        if jid in self.contacts:
            self.contacts[jid].update_subscription("subscribed", "")
        # Send the subscribed stanza to the server
        self.agent.client.send_presence(pto=jid, ptype="subscribed")

    # User-overridable methods
    def on_subscribe(self, peer_jid: str):
        pass

    def on_subscribed(self, peer_jid: str):
        pass

    def on_unsubscribe(self, peer_jid: str):
        pass

    def on_unsubscribed(self, peer_jid: str):
        pass

    def on_available(
        self,
        peer_jid: str,
        presence_info: PresenceInfo,
        last_presence: Optional[PresenceInfo],
    ):
        pass

    def on_unavailable(
        self,
        peer_jid: str,
        presence_info: PresenceInfo,
        last_presence: Optional[PresenceInfo],
    ):
        pass
